{-# OPTIONS -ffi #-}
{-# INCLUDE <event.h> #-}

-- TODOs:
--
-- 1) How and when to safely invoke freeFunPtr.
-- 2) Check whether the ForeignPtr to the event structures are the
--    right way of doing it.

module Event (Event, EventType(..), Event.init, create, add, del, dispatch) where

import Foreign
import Foreign.Ptr
import Foreign.C.Types
import System.Posix.Types
import Foreign.Marshal.Alloc

#include <event.h>

-- Possible event types enumeration
type CEventType = CShort

#{enum CEventType, ,
    ev_timeout = EV_TIMEOUT,
    ev_read    = EV_READ,
    ev_write   = EV_WRITE,
    ev_signal  = EV_SIGNAL,
    ev_persist = EV_PERSIST
}

data CEvent = CEvent -- corresponds to 'struct event'
type CEventPtr = Ptr CEvent
type CEventBasePtr = Ptr ()
type CTimeValPtr = Ptr () -- corresponds to 'struct timeval'

instance Storable CEvent where
    sizeOf _ = (#size struct event)
    alignment _ = alignment (undefined :: CInt)

-- Callback function types
type CCallback a = (CInt -> CShort -> Ptr a -> IO ())
type CCallbackPtr a = FunPtr (CCallback a)

-- libevent C-API

foreign import ccall "event_init"
    c_event_init :: IO CEventBasePtr

foreign import ccall "event_set"
    c_event_set :: CEventPtr -> CInt -> CShort -> CCallbackPtr a -> Ptr a -> IO ()

foreign import ccall "event_add"
    c_event_add :: CEventPtr -> CTimeValPtr -> IO CInt

foreign import ccall "event_del"
    c_event_del :: CEventPtr -> IO CInt

foreign import ccall "event_dispatch"
    c_event_dispatch :: IO CInt

-- Wrapper functionality

-- | Enumeration of all possible event types.
data EventType = TimeoutEvent | ReadEvent | WriteEvent | SignalEvent | PersistEvent
    deriving (Eq, Ord, Show)

toEventType :: CEventType -> EventType
toEventType e | e == ev_timeout = TimeoutEvent
              | e == ev_read    = ReadEvent
              | e == ev_write   = WriteEvent
              | e == ev_signal  = SignalEvent
              | e == ev_persist = PersistEvent

fromEventType :: EventType -> CEventType
fromEventType TimeoutEvent = ev_timeout
fromEventType ReadEvent    = ev_read
fromEventType WriteEvent   = ev_write
fromEventType SignalEvent  = ev_signal
fromEventType PersistEvent = ev_persist

-- Bit OR-ing the event types.
sumEventTypes :: [EventType] -> CEventType
sumEventTypes es = foldr (\e x -> e .|. x) 0 $ map fromEventType es

-- Create FunPtr from Haskell function, binding the add. argument in
-- a closure to be used during callback invocation.
mkCallback :: CallbackFn a -> a -> IO (CCallbackPtr a)
mkCallback f x = wrapCallback $ \i e _ -> f (Fd i) (toEventType e) x

foreign import ccall "wrapper"
    wrapCallback :: CCallback a -> IO (CCallbackPtr a)

-- | Datatype representing event tokens.
data Event = Event (ForeignPtr CEvent)

-- | Callback functions are invoked by the event dispatcher,
-- with the file descriptor causing the event, the event type
-- and a value given on event creation.
type CallbackFn a = (Fd -> EventType -> a -> IO ())

-- | Initializes the event system.
init :: IO ()
init = do
    c_event_init
    return ()

-- | Creates a new event for the given file descriptor and event types
-- which invokes the given callback function when triggered,
-- handling over the given argument.
create :: Fd -> [EventType] -> CallbackFn a -> a -> IO Event
create (Fd i) es fn x = do
    ep <- mallocForeignPtr
    cb <- mkCallback fn x
    withForeignPtr ep $ \eptr -> 
        c_event_set eptr (fromIntegral i) (sumEventTypes es) cb nullPtr
    return (Event ep)

-- | Adds then given event to the event loop, optionally taking a delay
-- value, when to make it available.
add :: Event -> IO Int
add (Event ep) = do
    withForeignPtr ep $ \p -> do
        r <- c_event_add p nullPtr
        return (fromIntegral r)

-- | Deletes the given event from the event loop.
del :: Event -> IO Int
del (Event ep) = do
    withForeignPtr ep $ \p -> do
        r <- c_event_del p
        return (fromIntegral r)

-- | Starts the event dispatching loop.
dispatch :: IO Int
dispatch = do
    r <- c_event_dispatch 
    return (fromIntegral r)

